package DBIx::SearchBuilder::Handle::mysqlPP;

use strict;
use warnings;

use base qw(DBIx::SearchBuilder::Handle::mysql);

1;

__END__

=head1 NAME

DBIx::SearchBuilder::Handle::mysqlPP - A mysql specific Handle object

=head1 DESCRIPTION

A Handle subclass for the "pure perl" mysql database driver.

This is currently identical to the DBIx::SearchBuilder::Handle::mysql class.
